#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cmath>
#include <cstdarg>
#include <cstdio>
#include <cstdlib>
#include <ctime>

#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <setjmp.h>

#include <iostream>
#include <new>

// Define CATCH_BAD_ALLOC_IN_CALLBACK to catch bad_alloc exceptions in
// callback function before returning to main.  Without this, we will
// try to use exception handling in the Fortran/foreign code, which
// causes trouble if it is not compiled with exception handling
// enabled.

#define CATCH_BAD_ALLOC_IN_CALLBACK

sigjmp_buf current_context;

sig_atomic_t interrupt_immediately = 0;

sig_atomic_t interrupt_state = 0;

sig_atomic_t allocation_error = 0;

int f77_exception_encountered = 0;

typedef void (*error_handler_t) (const char *, ...);

error_handler_t current_error_handler = 0;

extern "C" int
F77_FUNC (work, WORK) (int (*) (const int&), double *, const int&, const int&);

#define QUIT \
  do \
    { \
      if (interrupt_state) \
        { \
          interrupt_state = 0; \
          std::cerr << "do_something: throwing interrupt_exception" \
                    << std::endl; \
          throw interrupt_exception (); \
        } \
    } \
  while (0)

/* Hack to stringize macro results. */
#define xSTRINGIZE(x) #x
#define STRINGIZE(x) xSTRINGIZE(x)

/* How to print an error for the F77_XFCN macro. */

#define F77_XFCN_ERROR(f, F) \
  (*current_error_handler) \
    ("exception encountered in Fortran subroutine %s", \
     STRINGIZE (F77_FUNC (f, F)))

#define F77_XFCN(f, F, args) \
  do \
    { \
      sigjmp_buf saved_context; \
      f77_exception_encountered = 0; \
      save_current_context ((char *) saved_context); \
      if (sigsetjmp (current_context, 1)) \
	{ \
          restore_current_context ((char *) saved_context); \
	  if (f77_exception_encountered) \
	    F77_XFCN_ERROR (f, F); \
          else if (allocation_error) \
            { \
              throw bad_alloc (); \
            } \
          else \
            { \
              throw interrupt_exception (); \
            } \
	} \
      else \
        { \
          interrupt_immediately++; \
	  F77_FUNC (f, F) args; \
          interrupt_immediately--; \
          restore_current_context ((char *) saved_context); \
        } \
    } \
  while (0)

#define BEGIN_INTERRUPT_WITH_EXCEPTIONS \
  sig_atomic_t saved_interrupt_immediately = interrupt_immediately; \
 \
  try \
    { \
      interrupt_immediately = 0

#ifdef CATCH_BAD_ALLOC_IN_CALLBACK
#define CATCH_BAD_ALLOC \
  catch (bad_alloc) \
    { \
      std::cerr << "caught bad_alloc exception" << std::endl; \
      interrupt_immediately = saved_interrupt_immediately; \
      allocation_error = 1; \
      std::cerr << "jumping to enclosing context" << std::endl; \
      jump_to_enclosing_context (); \
    }
#else
#define CATCH_BAD_ALLOC
#endif

#define END_INTERRUPT_WITH_EXCEPTIONS \
    } \
  catch (interrupt_exception) \
    { \
      std::cerr << "caught interrupt exception" << std::endl; \
      interrupt_immediately = saved_interrupt_immediately; \
      std::cerr << "jumping to enclosing context" << std::endl; \
      jump_to_enclosing_context (); \
    } \
  CATCH_BAD_ALLOC \
 \
  interrupt_immediately = saved_interrupt_immediately

void
save_current_context (void *save_buf)
{
  memcpy (save_buf, current_context, sizeof (sigjmp_buf));
}

void
restore_current_context (void *save_buf)
{
  memcpy (current_context, save_buf, sizeof (sigjmp_buf));
}

void jump_to_enclosing_context (void) GCC_ATTR_NORETURN;

void
jump_to_enclosing_context (void)
{
  siglongjmp (current_context, 1);
}

class
interrupt_exception
{
  // Empty.
};

typedef RETSIGTYPE sig_handler (int);

sig_handler *
set_signal_handler (int sig, sig_handler *handler)
{
  struct sigaction act, oact;
  act.sa_handler = handler;
  act.sa_flags = 0;
  sigemptyset (&act.sa_mask);
  sigemptyset (&oact.sa_mask);
  sigaction (sig, &act, &oact);
  return oact.sa_handler;
}

RETSIGTYPE
sigint_handler (int)
{
  std::cerr << "sigint_handler" << std::endl;

  interrupt_state = 1;

  if (interrupt_immediately)
    jump_to_enclosing_context ();

  SIGHANDLER_RETURN (0);  
}

class
big_storage_hog
{
public:

  big_storage_hog (int n) : p (0)
  {
    std::cerr << "big_storage_hog: constructor" << std::endl;
    init (n);
  }

  ~big_storage_hog (void)
  {
    std::cerr << "big_storage_hog: destructor" << std::endl;
    delete [] p;
  }

  void init (int n)
  {
    if (n > 0)
      p = new double [n];
  }

  double *data (void) { return p; }

private:

  double *p;
};

static int
callback_fcn (const int& n)
{
  BEGIN_INTERRUPT_WITH_EXCEPTIONS;

  if (n > 0)
    {
      std::cerr << "callback_fcn: working for " << n << " seconds"
		<< std::endl;

      time_t t = time (0);
      double sum = 0.0;
      while (time (0) - t < n)
	{
	  QUIT;

	  for (int i = 0; i < 1000000; i++)
	    sum += i;
	}

      std::cerr << "callback_fcn: result is " << sum << std::endl;
    }
  else
    {
      // This allocation should fail, throwing bad_alloc exception.

      std::cerr << "callback_fcn: allocating 10GB" << std::endl;

      size_t mb = 1024 * 1024;
      size_t gb = 1024 * mb;
      char *t = new char [10*gb];

      std::cerr << "callback_fcn: initializing" << std::endl;

      for (size_t i = 0; i < 1024; i++)
	{
	  QUIT;

	  for (size_t j = 0; j < 10*mb; j++)
	    {
	      size_t k = j+1024*i;
	      t[k] = k;
	    }
	}

      std::cerr << "callback_fcn: done" << std::endl;
    }

  END_INTERRUPT_WITH_EXCEPTIONS;

  return 0;
}

static void
do_something (int k)
{
  if (k > 2)
    {
      std::cerr << "do_something: creating big_storage_hog object" << std::endl;
      int n = 1024 * 1024;
      big_storage_hog x (n);
      double *xp = x.data ();

      std::cerr << "do_something: calling work" << std::endl;

      F77_XFCN (work, WORK, (callback_fcn, xp, n, k));

      std::cerr << "do_something: done" << std::endl;
    }
  else
    {
      std::cerr << "do_something: calling work" << std::endl;

      F77_XFCN (work, WORK, (callback_fcn, 0, 0, k));

      std::cerr << "do_something: done" << std::endl;
    }

  QUIT;

  std::cerr << "do_something: returning" << std::endl;
}

// All the STOP statements in the Fortran routines have been replaced
// with a call to XSTOPX.
//
// XSTOPX jumps back to the entry point for the Fortran function that
// called us.  Then the calling function should do whatever cleanup
// is necessary.  */

extern "C" void
F77_FUNC (xstopx, XSTOPX) (const char *s, long int slen)
{
  f77_exception_encountered = 1;

  /* Skip printing message if it is just a single blank character.  */
  if (s && slen > 0 && ! (slen == 1 && *s == ' '))
    (*current_error_handler) ("%.*s\n", slen, s);

  jump_to_enclosing_context ();
}

void
error_handler (const char *fmt, ...)
{
  va_list args;
  va_start (args, fmt);
  vfprintf (stderr, fmt, args);
  va_end (args);
}

int
main (void)
{
  int k = 0;

  current_error_handler = error_handler;

  std::cerr << "main: setting SIGINT handler" << std::endl;

  set_signal_handler (SIGINT, sigint_handler);

  std::cerr << "main: entering main loop" << std::endl;

  while (1)
    {
      try
        {
          std::cerr << "main: calling do_something" << std::endl << std::flush;

          do_something (k);
        }
      catch (interrupt_exception)
        {
          std::cerr << "main: interrupt_exception caught" << std::endl;
	  interrupt_state = 0;
        }
      catch (bad_alloc)
	{
          std::cerr << "main: bad_alloc exception caught" << std::endl;
	  allocation_error = 0;
	}

      std::cerr << "k = " << k << std::endl;

      if (k++ > 2)
	break;
    }
}
